@@ -50,8 +50,46 @@ derived_probe::derived_probe (probe *p, probe_point *l):
     }
 }
 
-// ------------------------------------------------------------------------
 
+// ------------------------------------------------------------------------
+// Members of derived_probe_builder
+
+bool
+derived_probe_builder::get_param (std::map<std::string, literal*> const & params,
+                                  const std::string& key,
+                                  std::string& value)
+{
+  map<string, literal *>::const_iterator i = params.find (key);
+  if (i == params.end())
+    return false;
+  literal_string * ls = dynamic_cast<literal_string *>(i->second);
+  if (!ls)
+    return false;
+  value = ls->value;
+  return true;
+}
+
+
+bool
+derived_probe_builder::get_param (std::map<std::string, literal*> const & params,
+                                  const std::string& key,
+                                  int64_t& value)
+{
+  map<string, literal *>::const_iterator i = params.find (key);
+  if (i == params.end())
+    return false;
+  if (i->second == NULL)
+    return false;
+  literal_number * ln = dynamic_cast<literal_number *>(i->second);
+  if (!ln)
+    return false;
+  value = ln->value;
+  return true;
+}
+
+
+
+// ------------------------------------------------------------------------
 // Members of match_key.
 
 match_key::match_key(string const & n) 
@@ -173,18 +211,18 @@ match_node::find_builder(vector<probe_point::component *> const & components,
       // an entry in the sub table, and its value matches the rest
       // of the probe_point.
       match_key k(*components[pos]);
-      if (0) // session.verbose
+      if (0)
 	clog << "searching for component " << k.str() << endl;
       map<match_key, match_node *>::const_iterator i = sub.find(k);
       if (i == sub.end())
 	{
-	  if (0) // session.verbose
+	  if (0)
 	    clog << "no match found" << endl;
 	  return NULL;
 	}
       else
 	{
-	  if (0) // session.verbose
+	  if (0)
 	    clog << "matched " << k.str() << endl;
 	  derived_probe_builder * builder = NULL;
 	  if (k.have_parameter)
@@ -252,6 +290,7 @@ alias_expansion_builder
   
     // the token location of the use,
     n->tok = use->tok;
+    n->body->tok = use->tok;
 
     // and statements representing the concatenation of the alias'
     // body with the use's. 
@@ -373,6 +412,8 @@ symresolution_info::derive_probes (match_node * root,
           builder->build(session, p, loc, param_map, re_expand, dps);
           
           // Recursively expand any further-expanding results
+          // XXX: ... or the build routine could get a reference to
+          // this object and let it call derive_probes() internally
           if (!re_expand.empty())
             {
               for (unsigned j = 0; j < re_expand.size(); ++j)
@@ -396,53 +437,9 @@ symresolution_info::derive_probes (match_node * root,
 // Map usage checks
 //
 
-class lvalue_aware_traversing_visitor
-  : public traversing_visitor
-{
-  unsigned lval_depth;
-public:
-
-  lvalue_aware_traversing_visitor() : lval_depth(0) {}
-
-  bool is_in_lvalue()
-  {
-    return lval_depth > 0;
-  }
-
-  virtual void visit_pre_crement (pre_crement* e)
-  {
-    ++lval_depth;
-    e->operand->visit (this);
-    --lval_depth;
-  }
-
-  virtual void visit_post_crement (post_crement* e)
-  {
-    ++lval_depth;
-    e->operand->visit (this);
-    --lval_depth;
-  }
-  
-  virtual void visit_assignment (assignment* e)
-  {
-    ++lval_depth;
-    e->left->visit (this);
-    --lval_depth;
-    e->right->visit (this);
-  }
-
-  virtual void visit_delete_statement (delete_statement* s)
-  {
-    ++lval_depth;
-    s->value->visit (this);
-    --lval_depth;
-  }
-
-};
-
 
 struct mutated_map_collector
-  : public lvalue_aware_traversing_visitor
+  : public traversing_visitor
 {
   set<vardecl *> * mutated_maps;
 
@@ -452,14 +449,14 @@ struct mutated_map_collector
 
   void visit_arrayindex (arrayindex *e)
   {
-    if (is_in_lvalue())
+    if (is_active_lvalue(e))
       mutated_maps->insert(e->referent);
   }
 };
 
 
 struct no_map_mutation_during_iteration_check
-  : public lvalue_aware_traversing_visitor
+  : public traversing_visitor
 {
   systemtap_session & session;
   map<functiondecl *,set<vardecl *> *> & function_mutates_maps;
@@ -473,7 +470,7 @@ struct no_map_mutation_during_iteration_check
 
   void visit_arrayindex (arrayindex *e)
   {
-    if (is_in_lvalue())
+    if (is_active_lvalue(e))
       {
 	for (unsigned i = 0; i < maps_being_iterated.size(); ++i)
 	  {
@@ -730,6 +727,45 @@ symresolution_info::visit_foreach_loop (foreach_loop* e)
   e->block->visit (this);
 }
 
+struct 
+delete_statement_symresolution_info:
+  public traversing_visitor
+{
+  symresolution_info *parent;
+
+  delete_statement_symresolution_info (symresolution_info *p):
+    parent(p)
+  {}
+
+  void visit_arrayindex (arrayindex* e)
+  {
+    parent->visit_arrayindex (e);
+  }
+  void visit_functioncall (functioncall* e)
+  {
+    parent->visit_functioncall (e);
+  }
+
+  void visit_symbol (symbol* e)
+  {
+    if (e->referent)
+      return;
+    
+    vardecl* d = parent->find_var (e->name, -1);
+    if (d)
+      e->referent = d;
+    else
+      throw semantic_error ("unresolved array in delete statement", e->tok);
+  }
+};
+
+void 
+symresolution_info::visit_delete_statement (delete_statement* s)
+{
+  delete_statement_symresolution_info di (this);
+  s->value->visit (&di);
+}
+
 
 void
 symresolution_info::visit_symbol (symbol* e)
@@ -807,7 +843,7 @@ symresolution_info::visit_functioncall (functioncall* e)
 
 
 vardecl* 
-symresolution_info::find_var (const string& name, unsigned arity)
+symresolution_info::find_var (const string& name, int arity)
 {
 
   // search locals
@@ -837,12 +873,12 @@ symresolution_info::find_var (const string& name, unsigned arity)
   // search processed globals
   for (unsigned i=0; i<session.globals.size(); i++)
     if (session.globals[i]->name == name
-	&& session.globals[i]->compatible_arity(arity))
+	&& session.globals[i]->compatible_arity(arity))  
       {
 	session.globals[i]->set_arity (arity);
 	return session.globals[i];
       }
-
+  
   // search library globals
   for (unsigned i=0; i<session.library_files.size(); i++)
     {
@@ -852,7 +888,7 @@ symresolution_info::find_var (const string& name, unsigned arity)
           vardecl* g = f->globals[j];
           if (g->name == name && g->compatible_arity (arity))
             {
-              g->set_arity (arity);
+	      g->set_arity (arity);
               
               // put library into the queue if not already there	    
               if (find (session.files.begin(), session.files.end(), f) 
@@ -864,25 +900,6 @@ symresolution_info::find_var (const string& name, unsigned arity)
         }
     }
 
-  // search builtins that become locals
-  // XXX: need to invent a proper formalism for this
-  if (arity == 0 && (name == "$pid" || name == "$tid"))
-    {
-      vardecl_builtin* vb = new vardecl_builtin;
-      vb->name = name;
-      vb->type = pe_long;
-
-      // XXX: need a better way to synthesize tokens
-      token* t = new token;
-      t->type = tok_identifier;
-      t->content = name;
-      t->location.file = "<builtin>";
-      vb->tok = t;
-
-      locals.push_back (vb);
-      return vb;
-    }
-
   return 0;
 }
 
@@ -1257,7 +1274,7 @@ typeresolution_info::visit_ternary_expression (ternary_expression* e)
 
 template <class Referrer, class Referent>
 void resolve_2types (Referrer* referrer, Referent* referent,
-                    typeresolution_info* r, exp_type t)
+                    typeresolution_info* r, exp_type t, bool accept_unknown = false)
 {
   exp_type& re_type = referrer->type;
   const token* re_tok = referrer->tok;
@@ -1295,7 +1312,7 @@ void resolve_2types (Referrer* referrer, Referent* referent,
       r->resolved (te_tok, te_type);
       // catch re_type/t mismatch later
     }
-  else
+  else if (! accept_unknown)
     r->unresolved (re_tok);
 }
 
@@ -1307,12 +1324,18 @@ typeresolution_info::visit_symbol (symbol* e)
 
   if (e->referent->arity > 0)
     unresolved (e->tok); // symbol resolution should not permit this
-  // XXX: but consider "delete <array>;" and similar constructs
   else
     resolve_2types (e, e->referent, this, t);
 }
 
 
+void
+typeresolution_info::visit_target_symbol (target_symbol* e)
+{
+  throw semantic_error("unresolved target-symbol expression", e->tok);
+}
+
+
 void
 typeresolution_info::visit_arrayindex (arrayindex* e)
 {
@@ -1360,7 +1383,7 @@ typeresolution_info::visit_functioncall (functioncall* e)
 {
   assert (e->referent != 0);
 
-  resolve_2types (e, e->referent, this, t);
+  resolve_2types (e, e->referent, this, t, true); // accept unknown type 
 
   if (e->type == pe_stats)
     invalid (e->tok, e->type);
@@ -1510,11 +1533,35 @@ typeresolution_info::visit_expr_statement (expr_statement* e)
 }
 
 
+struct delete_statement_typeresolution_info: 
+  public throwing_visitor
+{
+  typeresolution_info *parent;
+  delete_statement_typeresolution_info (typeresolution_info *p):
+    throwing_visitor ("invalid operand of delete expression"),
+    parent (p)
+  {}
+
+  void visit_arrayindex (arrayindex* e)
+  {
+    parent->visit_arrayindex (e);
+  }
+  
+  void visit_symbol (symbol* e)
+  {
+    exp_type ignored = pe_unknown;
+    assert (e->referent != 0);    
+    resolve_2types (e, e->referent, parent, ignored);
+  }
+};
+
+
 void
 typeresolution_info::visit_delete_statement (delete_statement* e)
 {
-  // XXX: not yet supported
-  unresolved (e->tok);
+  delete_statement_typeresolution_info di (this);
+  t = pe_unknown;
+  e->value->visit (&di);
 }
 
 
