@@ -332,8 +332,32 @@ lexer::scan ()
 	  if (c == '\"') // closing double-quotes
 	    break;
 	  else if (c == '\\')
-	    {
-	      // XXX: handle escape sequences
+	    {	      
+	      c = input_get ();
+	      switch (c)
+		{
+		case 'a':
+		case 'b':
+		case 't':
+		case 'n':
+		case 'v':
+		case 'f':
+		case 'r':
+		case '\\':
+
+		  // Pass these escapes through to the string value
+		  // beign parsed; it will "likely" be emitted into 
+		  // a C literal. 
+		  //
+		  // XXX: verify this assumption.
+
+		  n->content.push_back('\\');
+
+		default:
+
+		  n->content.push_back(c);
+		  break;
+		}
 	    }
 	  else
 	    n->content.push_back(c);
@@ -750,6 +774,18 @@ parser::parse_functiondecl (std::vector<functiondecl*>& functions)
   fd->tok = t;
 
   t = next ();
+  if (t->type == tok_operator && t->content == ":")
+    {
+      t = next ();
+      if (t->type == tok_identifier && t->content == "string")
+	fd->type = pe_string;
+      else if (t->type == tok_identifier && t->content == "long")
+	fd->type = pe_long;
+      else throw parse_error ("expected 'string' or 'long'");
+
+      t = next ();
+    }
+
   if (! (t->type == tok_operator && t->content == "("))
     throw parse_error ("expected '('");
 
@@ -768,6 +804,17 @@ parser::parse_functiondecl (std::vector<functiondecl*>& functions)
       fd->formal_args.push_back (vd);
 
       t = next ();
+      if (t->type == tok_operator && t->content == ":")
+	{
+	  t = next ();
+	  if (t->type == tok_identifier && t->content == "string")
+	    vd->type = pe_string;
+	  else if (t->type == tok_identifier && t->content == "long")
+	    vd->type = pe_long;
+	  else throw parse_error ("expected 'string' or 'long'");
+	  
+	  t = next ();
+	}
       if (t->type == tok_operator && t->content == ")")
 	break;
       if (t->type == tok_operator && t->content == ",")
@@ -1626,19 +1673,12 @@ parser::parse_symbol ()
       while (true)
 	{
 	  string c;
-	  if (peek_op ("."))
-	    { 
-	      next();
-	      expect_ident (c);
-	      tsym->components.push_back
-		(make_pair (target_symbol::comp_struct_member, c));
-	    }
-	  else if (peek_op ("->"))
+	  if (peek_op ("->"))
 	    { 
 	      next(); 
 	      expect_ident (c);
 	      tsym->components.push_back
-		(make_pair (target_symbol::comp_struct_pointer_member, c));
+		(make_pair (target_symbol::comp_struct_member, c));
 	    }
 	  else if (peek_op ("["))
 	    { 
