@@ -35,9 +35,12 @@ int print_only = 0;
 int quiet = 0;
 int merge = 1;
 int verbose = 0;
+int enable_relayfs = 1;
+int target_pid = 0;
 unsigned int buffer_size = 0;
 char *modname = NULL;
 char *modpath = NULL;
+char *target_cmd = NULL;
 
  /* relayfs base file name */
 static char stpd_filebase[1024];
@@ -51,10 +54,14 @@ char *stp_check="stp_check";
 
 static void usage(char *prog)
 {
-	fprintf(stderr, "\n%s [-m] [-p] [-q] [-b bufsize] [-n num_subbufs] kmod-name\n", prog);
+	fprintf(stderr, "\n%s [-m] [-p] [-q] [-r] [-c cmd ] [-t pid] [-b bufsize] kmod-name\n", prog);
 	fprintf(stderr, "-m  Don't merge per-cpu files.\n");
 	fprintf(stderr, "-p  Print only.  Don't log to files.\n");
 	fprintf(stderr, "-q  Quiet. Don't display trace to stdout.\n");
+	fprintf(stderr, "-r  Disable running stp_check and loading relayfs module.\n");
+	fprintf(stderr, "-c cmd.  Command \'cmd\' will be run and stpd will exit when it does.\n");
+	fprintf(stderr, "   _stp_target will contain the pid for the command.\n");
+	fprintf(stderr, "-t pid.  Sets _stp_target to pid.\n");
 	fprintf(stderr, "-b buffer size. The systemtap module will specify a buffer size.\n");
 	fprintf(stderr, "   Setting one here will override that value. The value should be\n");
 	fprintf(stderr, "   an integer between 1 and 64 which be assumed to be the\n");
@@ -67,7 +74,7 @@ int main(int argc, char **argv)
 	int c, status;
 	pid_t pid;
 
-	while ((c = getopt(argc, argv, "mpqb:n:v")) != EOF) 
+	while ((c = getopt(argc, argv, "mpqrb:n:t:c:v")) != EOF) 
 	{
 		switch (c) {
 		case 'm':
@@ -82,6 +89,9 @@ int main(int argc, char **argv)
 		case 'v':
 			verbose = 1;
 			break;
+		case 'r':
+			enable_relayfs = 0;
+			break;
 		case 'b':
 		{
 			int size = (unsigned)atoi(optarg);
@@ -94,6 +104,12 @@ int main(int argc, char **argv)
 			buffer_size = size;
 			break;
 		}
+		case 't':
+			target_pid = atoi(optarg);
+			break;
+		case 'c':
+			target_cmd = optarg;
+			break;
 		default:
 			usage(argv[0]);
 		}
@@ -125,22 +141,25 @@ int main(int argc, char **argv)
 		usage(argv[0]);
 	}
 
-	/* now run the _stp_check script */
-	if ((pid = vfork()) < 0) {
-		perror ("vfork");
-		exit(-1);
-	} else if (pid == 0) {
-		if (execlp(stp_check, stp_check, NULL) < 0)
-			exit (-1);
-	}
-	if (waitpid(pid, &status, 0) < 0) {
-		perror("waitpid");
-		exit(-1);
-	}
-	if (WIFEXITED(status) && WEXITSTATUS(status)) {
-		perror (stp_check);
-		fprintf(stderr, "Could not execute %s\n", stp_check);
-		exit(1);
+
+	if (enable_relayfs) {
+		/* now run the _stp_check script */
+		if ((pid = vfork()) < 0) {
+			perror ("vfork");
+			exit(-1);
+		} else if (pid == 0) {
+			if (execlp(stp_check, stp_check, NULL) < 0)
+				exit (-1);
+		}
+		if (waitpid(pid, &status, 0) < 0) {
+			perror("waitpid");
+			exit(-1);
+		}
+		if (WIFEXITED(status) && WEXITSTATUS(status)) {
+			perror (stp_check);
+			fprintf(stderr, "Could not execute %s\n", stp_check);
+			exit(1);
+		}
 	}
 	
 	sprintf(stpd_filebase, "/mnt/relay/%d/cpu", getpid());
