@@ -38,6 +38,51 @@ extern "C" {
 
 using namespace std;
 
+
+// XXX: should standardize to these functions throughout translator
+
+template <typename OUT, typename IN> inline OUT
+lex_cast(IN const & in)
+{
+  stringstream ss;
+  OUT out;
+  if (!(ss << in && ss >> out))
+    throw runtime_error("bad lexical cast");
+  return out;
+}
+
+template <typename OUT, typename IN> inline OUT
+lex_cast_hex(IN const & in)
+{
+  stringstream ss;
+  OUT out;
+  if (!(ss << hex << showbase << in && ss >> out))
+    throw runtime_error("bad lexical cast");
+  return out;
+}
+
+
+// return as quoted string, with at least '"' backslash-escaped
+template <typename IN> inline string
+lex_cast_qstring(IN const & in)
+{
+  stringstream ss;
+  string out, out2;
+  if (!(ss << in))
+    throw runtime_error("bad lexical cast");
+  out = ss.str();
+  out2 += '"';
+  for (unsigned i=0; i<out.length(); i++)
+    {
+      if (out[i] == '"') // XXX others?
+	out2 += '\\';
+      out2 += out[i];
+    }
+  out2 += '"';
+  return out2;
+}
+
+
 // ------------------------------------------------------------------------
 // begin/end probes are run right during registration / deregistration
 // ------------------------------------------------------------------------
@@ -102,6 +147,8 @@ be_derived_probe::emit_probe_entries (translator_output* o, unsigned j)
       // While begin/end probes are executed single-threaded, we
       // still code defensively and use a per-cpu context.
       o->newline(1) << "struct context* c = & contexts [smp_processor_id()];";
+      o->newline() << "const char* probe_point = "
+		   << lex_cast_qstring(*l) << ";";
 
       // A precondition for running a probe handler is that we're in STARTING
       // or STOPPING state (not ERROR), and that no one else is already using
@@ -110,16 +157,15 @@ be_derived_probe::emit_probe_entries (translator_output* o, unsigned j)
       if (begin) o->line() << "STAP_SESSION_STARTING)";
       else o->line() << "STAP_SESSION_STOPPING)";
       o->newline(1) << "return;";
-      o->newline(-1) << "if (c->busy) {";
-      o->newline(1) << "printk (KERN_ERR \"probe reentrancy\");";
+      o->newline(-1) << "if (atomic_inc_return (&c->busy) != 1) {";
+      o->newline(1) << "printk (KERN_ERR \"probe reentrancy (%s vs %s)\\n\", "
+		    << "c->probe_point, probe_point);";
       o->newline() << "atomic_set (& session_state, STAP_SESSION_ERROR);";
       o->newline() << "return;";
       o->newline(-1) << "}";
       o->newline();
-
-      o->newline() << "c->busy ++;";
-      o->newline() << "mb ();"; // for smp
       o->newline() << "c->last_error = 0;";
+      o->newline() << "c->probe_point = probe_point;";
       o->newline() << "c->nesting = 0;";
       o->newline() << "c->regs = 0;";
       o->newline() << "c->actioncount = 0;";
@@ -132,8 +178,7 @@ be_derived_probe::emit_probe_entries (translator_output* o, unsigned j)
       o->newline() << "atomic_set (& session_state, STAP_SESSION_ERROR);";
       o->newline(-1) << "}";
 
-      o->newline() << "c->busy --;";
-      o->newline() << "mb ();";
+      o->newline() << "atomic_dec (&c->busy);";
       o->newline(-1) << "}" << endl;
     }
 }
@@ -156,32 +201,14 @@ static string TOK_LABEL("label");
 static string TOK_RELATIVE("relative");
 
 
-// XXX: should standardize to these functions throughout translator
 
-template <typename OUT, typename IN> inline OUT
-lex_cast(IN const & in)
-{
-  stringstream ss;
-  OUT out;
-  if (!(ss << in && ss >> out))
-    throw runtime_error("bad lexical cast");
-  return out;
-}
-
-template <typename OUT, typename IN> inline OUT
-lex_cast_hex(IN const & in)
-{
-  stringstream ss;
-  OUT out;
-  if (!(ss << hex << showbase << in && ss >> out))
-    throw runtime_error("bad lexical cast");
-  return out;
-}
 
 struct 
 func_info
 {
   string name;
+  char const * decl_file;
+  int decl_line;
   Dwarf_Die die;
   Dwarf_Addr prologue_end;
 };
@@ -190,6 +217,8 @@ struct
 inline_instance_info
 {
   string name;
+  char const * decl_file;
+  int decl_line;
   Dwarf_Die die;
 };
 
@@ -543,8 +572,8 @@ dwflpp
 				   void (* callback) (Dwarf_Line * line, void * arg),				   
 				   void *data)
   {
-    Dwarf_Line **srcsp;
-    size_t nsrcs;
+    Dwarf_Line **srcsp = NULL;
+    size_t nsrcs = 0;
 
     get_module_dwarf();
 
@@ -552,11 +581,19 @@ dwflpp
 		  dwarf_getsrc_file (module_dwarf, 
 				     srcfile, lineno, 0,
 				     &srcsp, &nsrcs));
-    
-    for (size_t i = 0; i < nsrcs; ++i)
+    try 
       {
-	callback (srcsp[i], data);
+	for (size_t i = 0; i < nsrcs; ++i)
+	  {
+	    callback (srcsp[i], data);
+	  }
       }
+    catch (...)
+      {
+	free (srcsp);
+	throw;
+      }    
+    free (srcsp);
   }
 
 
@@ -640,19 +677,25 @@ dwflpp
     return ( dwarf_lowpc (die, addr) == 0);
   }
 
-
-  char const * function_srcfile ()
-  {
-    assert (function);
-    return dwarf_func_file (function);
-  }
-
   void function_die (Dwarf_Die *d)
   {
     assert (function);
     dwarf_func_die (function, d);
   }
 
+  void function_file (char const ** c)
+  {
+    assert (function);
+    assert (c);
+    *c = dwarf_func_file (function);
+  }
+
+  void function_line (int *linep)
+  {
+    assert (function);
+    dwarf_func_line (function, linep);
+  }
+
   bool die_has_pc (Dwarf_Die * die, Dwarf_Addr pc)
   {
     int res = dwarf_haspc (die, pc);
@@ -682,7 +725,8 @@ dwflpp
 					       NULL, NULL, NULL, NULL));
   }
 
-  string literal_stmt_for_local(Dwarf_Addr pc,
+  string literal_stmt_for_local(Dwarf_Die *scope_die,
+				Dwarf_Addr pc,				
 				string const & local,
 				vector<pair<target_symbol::component_type,
 				std::string> > const & components,
@@ -692,8 +736,13 @@ dwflpp
 
     Dwarf_Die *scopes;
     Dwarf_Die vardie;
+    int nscopes = 0;
+
+    if (scope_die)
+      nscopes = dwarf_getscopes_die (scope_die, &scopes);
+    else
+      nscopes = dwarf_getscopes (cu, pc, &scopes);
 
-    int nscopes = dwarf_getscopes (cu, pc, &scopes);
     if (nscopes == 0)
       {
 	throw semantic_error ("unable to find any scopes containing "
@@ -892,11 +941,12 @@ dwflpp
     string prelude, postlude;
     switch (typetag)
       {
-
       default:
-	throw semantic_error ("target location not a base or pointer type");
+	throw semantic_error ("unsupported type tag "
+			      + lex_cast<string>(typetag));
 	break;
-	
+
+      case DW_TAG_enumeration_type:
       case DW_TAG_base_type:
 	ty = pe_long;
 	c_translate_fetch (&pool, 1, module_bias, die, typedie, &tail,
@@ -909,6 +959,7 @@ dwflpp
 	  Dwarf_Die pointee_typedie_mem;
 	  Dwarf_Die *pointee_typedie;
 	  Dwarf_Word pointee_encoding;
+	  Dwarf_Word pointee_byte_size = 0;
 	  int pointee_typetag;
 
 	  if (dwarf_attr_integrate (typedie, DW_AT_type, &attr_mem) == NULL)
@@ -930,44 +981,35 @@ dwflpp
 	      if (dwarf_attr_integrate (pointee_typedie, DW_AT_type, &attr_mem) == NULL)
 		throw semantic_error ("cannot get type of pointee: " + string(dwarf_errmsg (-1)));
 	    }
+
+	  if (dwarf_attr_integrate (pointee_typedie, DW_AT_byte_size, &attr_mem))
+	    dwarf_formudata (&attr_mem, &pointee_byte_size);
 	  
 	  dwarf_formudata (dwarf_attr_integrate (pointee_typedie, DW_AT_encoding, &attr_mem), 
 			   &pointee_encoding);
 
-	  if (pointee_typetag == DW_TAG_base_type
-	      && (pointee_encoding == DW_ATE_signed_char
-		  || pointee_encoding == DW_ATE_unsigned_char))
-	    {
-	      // We have an (un)signed char* or (un)signed char[], fetch it into 'tmpc' and
-	      // then copy to the return value via something strcpy-ish.
-	      ty = pe_string;
-	      prelude += "intptr_t tmpc = 0;\n";
-	      if (typetag == DW_TAG_array_type)
-		c_translate_array (&pool, 1, module_bias, typedie, &tail, NULL, 0);
-	      else
-		c_translate_pointer (&pool, 1, module_bias, typedie, &tail);
-	      c_translate_addressof (&pool, 1, module_bias, NULL, pointee_typedie, &tail, "tmpc");
-	      postlude += "deref_string (THIS->__retvalue, tmpc, MAXSTRINGLEN);\n";
-	    }
+	  // We have the pointer: cast it to an integral type via &(*(...))
+
+	  // NB: per bug #1187, at one point char*-like types were
+	  // automagically converted here to systemtap string values.
+	  // For several reasons, this was taken back out, leaving
+	  // pointer-to-string "conversion" (copying) to tapset functions.
+
+	  ty = pe_long;
+	  if (typetag == DW_TAG_array_type)
+	    c_translate_array (&pool, 1, module_bias, typedie, &tail, NULL, 0);
 	  else
-	    {
-	      // We have some other pointer: cast it to an integral type via &(*(...))
-	      ty = pe_long;
-	      if (typetag == DW_TAG_array_type)
-		c_translate_array (&pool, 1, module_bias, typedie, &tail, NULL, 0);
-	      else
-		c_translate_pointer (&pool, 1, module_bias, typedie, &tail);
-	      c_translate_addressof (&pool, 1, module_bias, NULL, pointee_typedie, &tail, 
-				     "THIS->__retvalue");
-	    }
+	    c_translate_pointer (&pool, 1, module_bias, typedie, &tail);
+	  c_translate_addressof (&pool, 1, module_bias, NULL, pointee_typedie, &tail, 
+				 "THIS->__retvalue");
 	}
 	break;	
       }
-
+    
     size_t bufsz = 1024;
     char *buf = static_cast<char*>(malloc(bufsz));
     assert(buf);
-
+    
     FILE *memstream = open_memstream (&buf, &bufsz);
     assert(memstream);
 
@@ -1018,12 +1060,13 @@ struct dwarf_query;
 
 struct dwarf_derived_probe : public derived_probe
 {
-  dwarf_derived_probe (dwarf_query & q,
-		       Dwarf_Addr addr);
+  dwarf_derived_probe (string const & funcname,
+		       char const * filename,
+		       int line,
+		       Dwarf_Die *scope_die,
+		       Dwarf_Addr addr,
+		       dwarf_query & q);
 
-  string module_name;
-  string function_name;
-  bool has_statement;
   Dwarf_Addr addr;
   Dwarf_Addr module_bias;
   bool has_return;
@@ -1302,7 +1345,12 @@ dwarf_query::parse_function_spec(string & spec)
 
 
 static void
-query_statement (Dwarf_Addr stmt_addr, dwarf_query * q)
+query_statement (string const & func,
+		 char const * file, 
+		 int line,
+		 Dwarf_Die *scope_die,
+		 Dwarf_Addr stmt_addr, 
+		 dwarf_query * q)
 {
   try
     {
@@ -1311,7 +1359,8 @@ query_statement (Dwarf_Addr stmt_addr, dwarf_query * q)
         throw semantic_error("incomplete: do not know how to interpret .relative",
                              q->base_probe->tok);
 
-      q->results.push_back(new dwarf_derived_probe(*q, stmt_addr));
+      q->results.push_back(new dwarf_derived_probe(func, file, line, 
+						   scope_die, stmt_addr, *q));
     }
   catch (const semantic_error& e)
     {
@@ -1321,7 +1370,7 @@ query_statement (Dwarf_Addr stmt_addr, dwarf_query * q)
 
 static void
 query_inline_instance_info (Dwarf_Addr entrypc,
-			    inline_instance_info const & ii,
+			    inline_instance_info & ii,
 			    dwarf_query * q)
 {
   if (q->has_return)
@@ -1334,13 +1383,14 @@ query_inline_instance_info (Dwarf_Addr entrypc,
 	clog << "querying entrypc " 
 	     << hex << entrypc << dec 
 	     << " of instance of inline '" << ii.name << "'" << endl;
-      query_statement (entrypc, q);
+      query_statement (ii.name, ii.decl_file, ii.decl_line, 
+		       &ii.die, entrypc, q);
     }
 }
 
 static void
 query_func_info (Dwarf_Addr entrypc,
-		 func_info const & fi,
+		 func_info & fi,
 		 dwarf_query * q)
 {
   if (q->has_return)
@@ -1348,14 +1398,18 @@ query_func_info (Dwarf_Addr entrypc,
       // NB. dwarf_derived_probe::emit_registrations will emit a
       // kretprobe based on the entrypc in this case.
       if (q->sess.verbose)
-	clog << "querying entrypc of function '" << fi.name << "' for return probe" << endl;
-      query_statement (entrypc, q);
+	clog << "querying entrypc of function '" 
+	     << fi.name << "' for return probe" << endl;
+      query_statement (fi.name, fi.decl_file, fi.decl_line, 
+		       &fi.die, entrypc, q);
     }
   else
     {
       if (q->sess.verbose)
-	clog << "querying prologue-end of function '" << fi.name << "'" << endl;
-      query_statement (fi.prologue_end, q);
+	clog << "querying prologue-end of function '" 
+	     << fi.name << "'" << endl;
+      query_statement (fi.name, fi.decl_file, fi.decl_line, 
+		       &fi.die, fi.prologue_end, q);
     }
 }
 
@@ -1367,7 +1421,7 @@ query_srcfile_line (Dwarf_Line * line, void * arg)
 
   Dwarf_Addr addr;
   dwarf_lineaddr(line, &addr);
-  
+
   for (map<Dwarf_Addr, func_info>::iterator i = q->filtered_functions.begin();
        i != q->filtered_functions.end(); ++i)
     {
@@ -1375,20 +1429,29 @@ query_srcfile_line (Dwarf_Line * line, void * arg)
 	{
 	  if (q->sess.verbose)
 	    clog << "function DIE lands on srcfile" << endl;
-	  query_func_info (i->first, i->second, q);
+	  if (q->has_statement_str)
+	    query_statement (i->second.name, i->second.decl_file, 
+			     q->line, NULL, addr, q);
+	  else
+	    query_func_info (i->first, i->second, q);
 	}
     }  
 
-  for (map<Dwarf_Addr, inline_instance_info>::iterator i = q->filtered_inlines.begin();
-       i != q->filtered_inlines.end(); ++i)
-    {
-      if (q->dw.die_has_pc (&(i->second.die), addr))
+      for (map<Dwarf_Addr, inline_instance_info>::iterator i 
+	     = q->filtered_inlines.begin();
+	   i != q->filtered_inlines.end(); ++i)
 	{
-	  if (q->sess.verbose)
-	    clog << "inline instance DIE lands on srcfile" << endl;
-	  query_inline_instance_info (i->first, i->second, q);
-	}
-    }  
+	  if (q->dw.die_has_pc (&(i->second.die), addr))
+	    {
+	      if (q->sess.verbose)
+		clog << "inline instance DIE lands on srcfile" << endl;
+	      if (q->has_statement_str)
+		query_statement (i->second.name, i->second.decl_file, 
+				 q->line, NULL, addr, q);
+	      else
+		query_inline_instance_info (i->first, i->second, q);
+	    }
+	}  
 }
 
 
@@ -1430,6 +1493,8 @@ query_dwarf_inline_instance (Dwarf_Die * die, void * arg)
 	      inline_instance_info inl;
 	      inl.die = *die;
 	      inl.name = q->dw.function_name;
+	      q->dw.function_file (&inl.decl_file);
+	      q->dw.function_line (&inl.decl_line);
 	      q->filtered_inlines[entrypc] = inl;
 	    }
 	}
@@ -1504,6 +1569,8 @@ query_dwarf_func (Dwarf_Func * func, void * arg)
 		  func_info func;
 		  q->dw.function_die (&func.die);
 		  func.name = q->dw.function_name;
+		  q->dw.function_file (&func.decl_file);
+		  q->dw.function_line (&func.decl_line);
 		  q->filtered_functions[entrypc] = func;
 		}
 	    }
@@ -1573,12 +1640,12 @@ query_cu (Dwarf_Die * cudie, void * arg)
 	  else
 	    {
 	      // Otherwise, simply probe all resolved functions...
-	      for (map<Dwarf_Addr, func_info>::const_iterator i = q->filtered_functions.begin();
+	      for (map<Dwarf_Addr, func_info>::iterator i = q->filtered_functions.begin();
 		   i != q->filtered_functions.end(); ++i)
 		query_func_info (i->first, i->second, q);
 
 	      // And all inline instances.
-	      for (map<Dwarf_Addr, inline_instance_info>::const_iterator i 
+	      for (map<Dwarf_Addr, inline_instance_info>::iterator i 
 		     = q->filtered_inlines.begin(); i != q->filtered_inlines.end(); ++i)
 		query_inline_instance_info (i->first, i->second, q);
 
@@ -1594,7 +1661,7 @@ query_cu (Dwarf_Die * cudie, void * arg)
 	  if (q->has_module)
 	    query_addr = q->dw.module_address_to_global(query_addr);
 
-	  query_statement (query_addr, q);
+	  query_statement ("", "", -1, NULL, query_addr, q);
         }
       return DWARF_CB_OK;
     }
@@ -1686,10 +1753,11 @@ var_expanding_copy_visitor
   static unsigned tick;
 
   dwarf_query & q;
+  Dwarf_Die *scope_die;
   Dwarf_Addr addr;
 
-  var_expanding_copy_visitor(dwarf_query & q, Dwarf_Addr a)
-    : q(q), addr(a)
+  var_expanding_copy_visitor(dwarf_query & q, Dwarf_Die *sd, Dwarf_Addr a)
+    : q(q), scope_die(sd), addr(a)
   {}
   void visit_target_symbol (target_symbol* e);
 };
@@ -1715,10 +1783,20 @@ var_expanding_copy_visitor::visit_target_symbol (target_symbol *e)
   functiondecl *fdecl = new functiondecl;
   embeddedcode *ec = new embeddedcode;
   ec->tok = e->tok;
-  ec->code = q.dw.literal_stmt_for_local(addr,
-					 e->base_name.substr(1),
-					 e->components,
-					 fdecl->type);
+  try
+    {      
+      ec->code = q.dw.literal_stmt_for_local(scope_die, 
+					     addr,
+					     e->base_name.substr(1),
+					     e->components,
+					     fdecl->type);
+    }
+  catch (const semantic_error& er)
+    {
+      semantic_error er2 (er);
+      er2.tok1 = e->tok;
+      throw er2;
+    }
   fdecl->name = fname;
   fdecl->body = ec;
   q.sess.functions.push_back(fdecl);
@@ -1732,16 +1810,18 @@ var_expanding_copy_visitor::visit_target_symbol (target_symbol *e)
 }
 
 
-dwarf_derived_probe::dwarf_derived_probe (dwarf_query & q,
-					  Dwarf_Addr addr)
+dwarf_derived_probe::dwarf_derived_probe (string const & funcname,
+					  char const * filename,
+					  int line,
+					  Dwarf_Die *scope_die,
+					  Dwarf_Addr addr,
+					  dwarf_query & q)
   : derived_probe (NULL),
-    module_name(q.dw.module_name),
-    function_name(q.dw.function_name),
-    has_statement(q.has_statement_str || q.has_statement_num),
     addr(addr),
     module_bias(q.dw.module_bias),
     has_return (q.has_return)
 {
+  string module_name(q.dw.module_name);
   // Lock the kernel module in memory.
   if (module_name != TOK_KERNEL)
     {
@@ -1773,16 +1853,11 @@ dwarf_derived_probe::dwarf_derived_probe (dwarf_query & q,
 
   if (q.has_function_str || q.has_statement_str)
       {
-        string retro_name;;
-        if (! function_name.empty())
-          retro_name = function_name + "@" + q.dw.cu_name; // XXX: add line number
-        else if (q.has_function_str)
-          retro_name = q.function_str_val;
-        else // has_statement_str
-          retro_name = q.statement_str_val;
-        // XXX: actually the statement_str case is not yet adequately
-        // handled in the search code
-
+        string retro_name = funcname;
+	if (filename && !string (filename).empty())
+	  retro_name += ("@" + string (filename));
+	if (line != -1)
+	  retro_name += (":" + lex_cast<string> (line));
         comps.push_back
           (new probe_point::component
            (fn_or_stmt, new literal_string (retro_name)));
@@ -1808,7 +1883,7 @@ dwarf_derived_probe::dwarf_derived_probe (dwarf_query & q,
   locations.push_back(new probe_point(comps, q.base_probe->locations[0]->tok));
 
   // Now make a local-variable-expanded copy of the probe body
-  var_expanding_copy_visitor v (q, addr);
+  var_expanding_copy_visitor v (q, scope_die, addr);
   require <block*> (&v, &(this->body), q.base_probe->body);
   this->tok = q.base_probe->tok;
 }
@@ -1943,6 +2018,29 @@ dwarf_derived_probe::emit_deregistrations (translator_output* o, unsigned proben
 void
 dwarf_derived_probe::emit_probe_entries (translator_output* o, unsigned probenum)
 {
+  static unsigned already_emitted_fault_handler = 0;
+
+  if (! already_emitted_fault_handler)
+    {
+      o->newline() << "int stap_kprobe_fault_handler (struct kprobe* kp, "
+		   << "struct pt_regs* regs, int trapnr) {";
+      o->newline(1) << "struct context *c = & contexts [smp_processor_id()];";
+      o->newline() << "printk (KERN_ERR \"systemtap probe fault\\n\");";
+      o->newline() << "printk (KERN_ERR \"cpu %d, probe %s, near %s\\n\", ";
+      o->newline(1) << "smp_processor_id(), ";
+      o->newline() << "c->probe_point ? c->probe_point : \"unknown\", ";
+      o->newline() << "c->last_stmt ? c->last_stmt : \"unknown\");";
+      o->newline() << "c->last_error = \"probe faulted\";";
+      o->newline(-1) << "atomic_set (& session_state, STAP_SESSION_ERROR);";
+
+      o->newline() << "return 0;"; // defer to kernel fault handler
+      // NB: We might prefer to use "return 1" instead, to consider
+      // the fault "handled".  But we may get into an infinite loop
+      // of traps if the faulting instruction is simply restarted.
+
+      o->newline(-1) << "}";
+      already_emitted_fault_handler ++;
+    }
 
   // Construct a single entry function, and a struct kprobe pointing into
   // the entry function. The entry function will call the probe function.
@@ -1955,22 +2053,23 @@ dwarf_derived_probe::emit_probe_entries (translator_output* o, unsigned probenum
     o->line() << "struct kprobe *_ignored";
   o->line() << ", struct pt_regs *regs) {";
   o->newline(1) << "struct context *c = & contexts [smp_processor_id()];";
-  o->newline();
+  o->newline() << "const char* probe_point = "
+	       << lex_cast_qstring(*locations[0]) << ";";
 
   // A precondition for running a probe handler is that we're in RUNNING
   // state (not ERROR), and that no one else is already using this context.
   o->newline() << "if (atomic_read (&session_state) != STAP_SESSION_RUNNING)";
   o->newline(1) << "return 0;";
-  o->newline(-1) << "if (c->busy) {";
-  o->newline(1) << "printk (KERN_ERR \"probe reentrancy\");";
+  o->newline(-1) << "if (atomic_inc_return (&c->busy) != 1) {";
+  o->newline(1) << "printk (KERN_ERR \"probe reentrancy (%s vs %s)\\n\", "
+		    << "c->probe_point, probe_point);";
   o->newline() << "atomic_set (& session_state, STAP_SESSION_ERROR);";
   o->newline() << "return 0;";
   o->newline(-1) << "}";
   o->newline();
 
-  o->newline() << "c->busy ++;";
-  o->newline() << "mb ();"; // for smp
   o->newline() << "c->last_error = 0;";
+  o->newline() << "c->probe_point = probe_point;";
   o->newline() << "c->nesting = 0;";
   o->newline() << "c->regs = regs;";
   o->newline() << "c->actioncount = 0;";
@@ -1983,9 +2082,7 @@ dwarf_derived_probe::emit_probe_entries (translator_output* o, unsigned probenum
   o->newline() << "atomic_set (& session_state, STAP_SESSION_ERROR);";
   o->newline(-1) << "}";
   
-  o->newline() << "c->busy --;";
-  o->newline() << "mb ();";
-
+  o->newline() << "atomic_dec (& c->busy);";
   o->newline() << "return 0;";
   o->newline(-1) << "}" << endl;
 
@@ -1998,6 +2095,8 @@ dwarf_derived_probe::emit_probe_entries (translator_output* o, unsigned probenum
       o->newline(1) << ".kp.addr = 0," ;
       o->newline() << ".handler = &"
                    << probe_entry_function_name(probenum) << ",";
+      // XXX: pending PR 1289
+      // o->newline() << ".kp.fault_handler = &stap_kprobe_fault_handler,";
       o->newline() << ".maxactive = 1";
       o->newline(-1) << "};";
     }
@@ -2007,6 +2106,8 @@ dwarf_derived_probe::emit_probe_entries (translator_output* o, unsigned probenum
                    << probe_entry_struct_kprobe_name(probenum)
                    << "= {";
       o->newline(1) << ".addr       = 0," ;
+      // XXX: pending PR 1289
+      // o->newline() << ".fault_handler = &stap_kprobe_fault_handler,";
       o->newline() << ".pre_handler = &" << probe_entry_function_name(probenum);
       o->newline(-1) << "};";
     }
@@ -2110,7 +2211,8 @@ timer_derived_probe::emit_probe_entries (translator_output* o, unsigned j)
 
   o->newline() << "void enter_" << j << " (unsigned long val) {";
   o->newline(1) << "struct context* c = & contexts [smp_processor_id()];";
-  
+  o->newline() << "const char* probe_point = "
+	       << lex_cast_qstring(*locations[0]) << ";";
   o->newline() << "(void) val;";
 
   // A precondition for running a probe handler is that we're in
@@ -2119,8 +2221,9 @@ timer_derived_probe::emit_probe_entries (translator_output* o, unsigned j)
   o->newline() << "if (atomic_read (&session_state) != STAP_SESSION_RUNNING)";
   o->newline(1) << "return;";
 
-  o->newline(-1) << "if (c->busy) {";
-  o->newline(1) << "printk (KERN_ERR \"probe reentrancy\");";
+  o->newline(-1) << "if (atomic_inc_return (&c->busy) != 1) {";
+  o->newline(1) << "printk (KERN_ERR \"probe reentrancy (%s vs %s)\\n\", "
+		<< "c->probe_point, probe_point);";
   o->newline() << "atomic_set (& session_state, STAP_SESSION_ERROR);";
   o->newline() << "return;";
   o->newline(-1) << "}";
@@ -2132,15 +2235,16 @@ timer_derived_probe::emit_probe_entries (translator_output* o, unsigned j)
     o->line() << " + _stp_random_pm(" << randomize << ")";
   o->line() << ");";
   
-  o->newline() << "c->busy ++;";
-  o->newline() << "mb ();"; // for smp
+  o->newline() << "c->probe_point = probe_point;";
   o->newline() << "c->last_error = 0;";
   o->newline() << "c->nesting = 0;";
+  o->newline() << "c->regs = 0;";
+
+  o->newline() << "#ifdef __i386__";   // task_pt_regs is i386-only
   o->newline() << "if (! in_interrupt())";
-  o->newline(1) << "c->regs = 0;";
-  o->newline(-1) << "else";
   o->newline(1) << "c->regs = task_pt_regs (current);";
-  o->indent(-1);
+  o->newline(-1) << "#endif";
+
   o->newline() << "c->actioncount = 0;";
   
   // NB: locals are initialized by probe function itself
@@ -2151,8 +2255,7 @@ timer_derived_probe::emit_probe_entries (translator_output* o, unsigned j)
   o->newline() << "atomic_set (& session_state, STAP_SESSION_ERROR);";
   o->newline(-1) << "}";
   
-  o->newline() << "c->busy --;";
-  o->newline() << "mb ();";
+  o->newline() << "atomic_dec (&c->busy);";
   o->newline(-1) << "}" << endl;
 }
 
