@@ -30,6 +30,27 @@ stringify(T t)
   return s.str ();
 }
 
+// return as quoted string, with at least '"' backslash-escaped
+template <typename IN> inline string
+lex_cast_qstring(IN const & in)
+{
+  stringstream ss;
+  string out, out2;
+  if (!(ss << in))
+    throw runtime_error("bad lexical cast");
+  out = ss.str();
+  out2 += '"';
+  for (unsigned i=0; i<out.length(); i++)
+    {
+      if (out[i] == '"') // XXX others?
+	out2 += '\\';
+      out2 += out[i];
+    }
+  out2 += '"';
+  return out2;
+}
+
+
 struct var;
 struct tmpvar;
 struct mapvar;
@@ -281,12 +302,40 @@ struct varlock
   c_unparser & c;
   var const & v;
   bool w;
+  string post_unlock_label;
 
   varlock(c_unparser & c, var const & v, bool w): c(c), v(v), w(w)
   {
     if (v.is_local()) return;
-    c.o->newline() << (w ? "write_lock" : "read_lock")
-                   << " (& " << v << "_lock);";
+
+    if (!w) // read lock - no need to try, just do it
+      c.o->newline() << "read_lock (& " << v << "_lock);";
+    else
+      {
+	// There may be a few opportunities for deadlock beyond bug #1268
+	// (foreach).  In general, we use a loop with FOO_trylock with a
+	// manual "timeout".  If it takes too many iterations to acquire
+	// the lock, we signal a deadlock error in the context and jump
+	// over all the code, right past the corresponding unlock.
+	
+	static unsigned unlock_label_counter = 0;
+	post_unlock_label = string ("post_unlock_")
+	  + stringify (unlock_label_counter ++);
+	
+	c.o->newline() << "{";
+	c.o->newline(1) << "unsigned trylock_count = 0;";
+	c.o->newline() << "while ("
+		       << "!(write_trylock (& " << v << "_lock))"
+		       << " && "
+		       << "(trylock_count++ < MAXTRYLOCK)"
+		       << ") ; /* spin */";
+	c.o->newline() << "if (unlikely (trylock_count >= MAXTRYLOCK)) {";
+	c.o->newline(1) << "c->last_error = \"deadlock over variable "
+			<< v << "\";";
+	c.o->newline() << "goto " << post_unlock_label << ";";
+	c.o->newline(-1) << "}";
+	c.o->newline(-1) << "}";
+      }
   }
 
   ~varlock()
@@ -294,6 +343,11 @@ struct varlock
     if (v.is_local()) return;
     c.o->newline() << (w ? "write_unlock" : "read_unlock")
                    << " (& " << v << "_lock);";
+    if (w)
+      {
+	c.o->newline(-1) << post_unlock_label << ": ;";
+	c.o->indent(1);
+      }
   }
 };
 
@@ -527,6 +581,7 @@ c_unparser::emit_common_header ()
   o->newline() << "#define MAXNESTING 30";
   o->newline() << "#define MAXCONCURRENCY NR_CPUS";
   o->newline() << "#define MAXSTRINGLEN 128";
+  o->newline() << "#define MAXTRYLOCK 20";
   o->newline() << "#define MAXACTION 1000";
   o->newline() << "#define MAXMAPENTRIES 2048";
   o->newline();
@@ -541,7 +596,8 @@ c_unparser::emit_common_header ()
   o->newline() << "atomic_t session_state = ATOMIC_INIT (STAP_SESSION_STARTING);";
   o->newline();
   o->newline() << "struct context {";
-  o->newline(1) << "unsigned busy;"; // XXX: should be atomic_t ?
+  o->newline(1) << "atomic_t busy;";
+  o->newline() << "const char *probe_point;";
   o->newline() << "unsigned actioncount;";
   o->newline() << "unsigned nesting;";
   o->newline() << "const char *last_error;";
@@ -564,8 +620,15 @@ c_unparser::emit_common_header ()
       for (unsigned j=0; j<dp->locals.size(); j++)
         {
           vardecl* v = dp->locals[j];
-          o->newline() << c_typename (v->type) << " " 
-		       << c_varname (v->name) << ";";
+	  try 
+	    {
+	      o->newline() << c_typename (v->type) << " " 
+			   << c_varname (v->name) << ";";
+	    } catch (const semantic_error& e) {
+	      semantic_error e2 (e);
+	      if (e2.tok1 == 0) e2.tok1 = v->tok;
+	      throw e2;
+	    }
         }
       c_tmpcounter ct (this);
       dp->body->visit (& ct);
@@ -580,15 +643,29 @@ c_unparser::emit_common_header ()
       o->indent(1);
       for (unsigned j=0; j<fd->locals.size(); j++)
         {
-          vardecl* v = fd->locals[j];
-          o->newline() << c_typename (v->type) << " " 
-  		       << c_varname (v->name) << ";";
+	  vardecl* v = fd->locals[j];
+	  try 
+	    {
+	      o->newline() << c_typename (v->type) << " " 
+			   << c_varname (v->name) << ";";
+	    } catch (const semantic_error& e) {
+	      semantic_error e2 (e);
+	      if (e2.tok1 == 0) e2.tok1 = v->tok;
+	      throw e2;
+	    }
         }
       for (unsigned j=0; j<fd->formal_args.size(); j++)
         {
           vardecl* v = fd->formal_args[j];
-          o->newline() << c_typename (v->type) << " " 
-		       << c_varname (v->name) << ";";
+	  try 
+	    {
+	      o->newline() << c_typename (v->type) << " " 
+			   << c_varname (v->name) << ";";
+	    } catch (const semantic_error& e) {
+	      semantic_error e2 (e);
+	      if (e2.tok1 == 0) e2.tok1 = v->tok;
+	      throw e2;
+	    }
         }
       c_tmpcounter ct (this);
       fd->body->visit (& ct);
@@ -657,6 +734,8 @@ c_unparser::emit_module_init ()
   for (unsigned i=0; i<session->probes.size(); i++)
     {
       o->newline() << "/* register " << i << " */";
+      for (unsigned k=0; k<session->probes[i]->locations.size(); k++)
+	o->newline() << "/* " << *session->probes[i]->locations[k] << " */";
       session->probes[i]->emit_registrations (o, i);
 
       o->newline() << "if (unlikely (rc)) {";
@@ -722,14 +801,13 @@ c_unparser::emit_module_exit ()
   o->newline() << "do {";
   o->newline(1) << "int i;";
   o->newline() << "holdon = 0;";
-  o->newline() << "mb ();";
   o->newline() << "for (i=0; i<NR_CPUS; i++)";
-  o->newline(1) << "if (contexts[i].busy) holdon = 1;";
+  o->newline(1) << "if (atomic_read (&contexts[i].busy)) holdon = 1;";
   // o->newline(-1) << "if (holdon) msleep (5);";
   o->newline(-1) << "} while (holdon);";
   o->newline(-1);
   // XXX: might like to have an escape hatch, in case some probe is
-  // genuinely stuck
+  // genuinely stuck somehow
 
   for (int i=session->probes.size()-1; i>=0; i--)
     session->probes[i]->emit_deregistrations (o, i);
@@ -740,7 +818,6 @@ c_unparser::emit_module_exit ()
       if (v->index_types.size() > 0)
 	o->newline() << getmap (v).fini();
     }
-  // XXX: if anyrc, log badness
   o->newline(-1) << "}" << endl;
 }
 
@@ -772,7 +849,7 @@ c_unparser::emit_function (functiondecl* v)
   for (unsigned i=0; i<v->locals.size(); i++)
     {
       if (v->locals[i]->index_types.size() > 0) // array?
-	throw semantic_error ("array locals not supported", v->tok);
+	throw semantic_error ("array locals not supported", v->locals[i]->tok);
 
       o->newline() << getvar (v->locals[i]).init();
     }
@@ -834,10 +911,8 @@ c_unparser::emit_probe (derived_probe* v, unsigned i)
   this->current_probenum = 0; // not essential
 
   o->newline(-1) << "out:";
-  o->newline(1) << ";";
-  
-  // XXX: uninitialize locals
-
+  // NB: no need to uninitialize locals, except if arrays can somedays be local
+  o->newline(1) << "_stp_print_flush();";
   o->newline(-1) << "}" << endl;
   
   v->emit_probe_entries (o, i);
@@ -1159,7 +1234,7 @@ c_unparser::c_declare_static(exp_type ty, const string &name)
 void 
 c_unparser::c_strcpy (const string& lvalue, const string& rvalue) 
 {
-  o->newline() << "strncpy (" 
+  o->newline() << "strlcpy (" 
 		   << lvalue << ", " 
 		   << rvalue << ", MAXSTRINGLEN);";
 }
@@ -1168,7 +1243,7 @@ c_unparser::c_strcpy (const string& lvalue, const string& rvalue)
 void 
 c_unparser::c_strcpy (const string& lvalue, expression* rvalue) 
 {
-  o->newline() << "strncpy (" << lvalue << ", ";
+  o->newline() << "strlcpy (" << lvalue << ", ";
   rvalue->visit (this);
   o->line() << ", MAXSTRINGLEN);";
 }
@@ -1177,16 +1252,16 @@ c_unparser::c_strcpy (const string& lvalue, expression* rvalue)
 void 
 c_unparser::c_strcat (const string& lvalue, const string& rvalue) 
 {
-  o->newline() << "strncat (" 
-		   << lvalue << ", " 
-		   << rvalue << ", MAXSTRINGLEN);";
+  o->newline() << "strlcat (" 
+	       << lvalue << ", " 
+	       << rvalue << ", MAXSTRINGLEN);";
 }
 
 
 void 
 c_unparser::c_strcat (const string& lvalue, expression* rvalue) 
 {
-  o->newline() << "strncat (" << lvalue << ", ";
+  o->newline() << "strlcat (" << lvalue << ", ";
   rvalue->visit (this);
   o->line() << ", MAXSTRINGLEN);";
 }
@@ -1280,7 +1355,7 @@ c_unparser::visit_statement (statement *s, unsigned actions)
 
   o->newline() << "if (unlikely (c->last_error)) goto " << outlabel << ";";
   assert (s->tok);
-  o->newline() << "c->last_stmt = \"" << *s->tok << "\";";
+  o->newline() << "c->last_stmt = " << lex_cast_qstring(*s->tok) << ";";
   if (actions > 0)
     {
       o->newline() << "c->actioncount += " << actions << ";";
@@ -1440,7 +1515,13 @@ c_unparser::visit_foreach_loop (foreach_loop *s)
   // NB: structure parallels for_loop
 
   // initialization
-  varlock_r guard (*this, mv);
+
+  // XXX: until bug #1275 is fixed, any reference into a foreach()-iterated
+  // array is an instant deadlock.  For now, don't lock around foreach(),
+  // and hope that no concurrent probe handler will modify this array.
+  //
+  // varlock_r guard (*this, mv);
+
   o->newline() << iv << " = " << iv.start (mv) << ";";
 
   // condition
@@ -1470,7 +1551,7 @@ c_unparser::visit_foreach_loop (foreach_loop *s)
     
   // exit
   o->newline(-1) << breaklabel << ":";
-  o->indent(1);
+  o->newline(1) << "; /* dummy statement */";
   // varlock dtor will show up here
 }
 
@@ -1579,7 +1660,14 @@ c_unparser::visit_continue_statement (continue_statement* s)
 void
 c_unparser::visit_literal_string (literal_string* e)
 {
-  o->line() << '"' << e->value << '"'; // XXX: escape special chars
+  const string& v = e->value;
+  o->line() << '"';
+  for (unsigned i=0; i<v.size(); i++)
+    if (v[i] == '"') // or other escapeworthy characters?
+      o->line() << '\\' << '"';
+    else
+      o->line() << v[i];
+  o->line() << '"';
 }
 
 
@@ -1588,7 +1676,8 @@ c_unparser::visit_literal_number (literal_number* e)
 {
   // This looks ugly, but tries to be warning-free on 32- and 64-bit
   // hosts.
-  o->line() << "((uint64_t)" << e->value << "LL)";
+  // NB: this needs to be signed!
+  o->line() << "((int64_t)" << e->value << "LL)";
 }
 
 
@@ -1640,7 +1729,7 @@ c_unparser::visit_binary_expression (binary_expression* e)
       tmpvar right = gensym (pe_long);
 
       o->line() << "({";
-
+      o->newline() << "c->last_stmt = " << lex_cast_qstring(*e->tok) << ";";
       o->newline(1) << left << " = ";
       e->left->visit (this);
       o->line() << ";";
@@ -1729,12 +1818,14 @@ c_unparser::visit_array_in (array_in* e)
 
   vector<tmpvar> idx;
   load_map_indices (e->operand, idx);
+  o->newline() << "c->last_stmt = " << lex_cast_qstring(*e->tok) << ";";
 
   tmpvar res = gensym (pe_long);
 
   { // block used to control varlock_r lifespan
     mapvar mvar = getmap (e->operand->referent, e->tok);
-    varlock_r guard (*this, mvar);
+    // XXX: should be varlock_r, but runtime arrays reads mutate
+    varlock_w guard (*this, mvar);
     o->newline() << mvar.seek (idx) << ";";
     c_assign (res, mvar.exists(), e->tok);
   }
@@ -1805,6 +1896,7 @@ c_unparser::visit_concatenation (concatenation* e)
   
   o->line() << "({ ";
   o->indent(1);
+  o->newline() << "c->last_stmt = " << lex_cast_qstring(*e->tok) << ";";
   c_assign (t.qname(), e->left, "assignment");
   c_strcat (t.qname(), e->right);
   o->newline() << t << ";";
@@ -1904,10 +1996,6 @@ c_unparser::visit_symbol (symbol* e)
   if (r->index_types.size() != 0)
     throw semantic_error ("invalid reference to array", e->tok);
 
-  // XXX: handle special macro symbols
-  // XXX: acquire read lock on global; copy value
-  // into local temporary
-
   var v = getvar(r, e->tok);
   o->line() << v;
 }
@@ -1975,6 +2063,7 @@ c_unparser_assignment::visit_symbol (symbol *e)
   if (e->referent->index_types.size() != 0)
     throw semantic_error ("unexpected reference to array", e->tok);
 
+  parent->o->newline() << "c->last_stmt = " << lex_cast_qstring(*e->tok) << ";";
   tmpvar rval = parent->gensym (e->type);
   tmpvar res = parent->gensym (e->type);
 
@@ -2034,6 +2123,8 @@ c_unparser::load_map_indices(arrayindex *e,
 	throw semantic_error ("array index type mismatch", e->indexes[i]->tok);
       
       tmpvar ix = gensym (r->index_types[i]);
+      o->newline() << "c->last_stmt = "
+		   << lex_cast_qstring(*e->indexes[i]->tok) << ";";
       c_assign (ix.qname(), e->indexes[i], "array index copy");
       idx.push_back (ix);
     }
@@ -2068,7 +2159,9 @@ c_unparser::visit_arrayindex (arrayindex* e)
   
   { // block used to control varlock_r lifespan
     mapvar mvar = getmap (e->referent, e->tok);
-    varlock_r guard (*this, mvar);
+    // XXX: should be varlock_r, but runtime arrays reads mutate
+    o->newline() << "c->last_stmt = " << lex_cast_qstring(*e->tok) << ";";
+    varlock_w guard (*this, mvar);
     o->newline() << mvar.seek (idx) << ";";
     c_assign (res, mvar.get(), e->tok);
   }
@@ -2144,6 +2237,7 @@ c_unparser_assignment::visit_arrayindex (arrayindex *e)
   
   { // block used to control varlock_w lifespan
     mapvar mvar = parent->getmap (e->referent, e->tok);
+    o->newline() << "c->last_stmt = " << lex_cast_qstring(*e->tok) << ";";
     varlock_w guard (*parent, mvar);
     o->newline() << mvar.seek (idx) << ";";
     if (op != "=") // don't bother fetch slot if we will just overwrite it
@@ -2198,13 +2292,16 @@ c_unparser::visit_functioncall (functioncall* e)
 	throw semantic_error ("function argument type mismatch",
 			      e->args[i]->tok, "vs", r->formal_args[i]->tok);
 
+      o->newline() << "c->last_stmt = "
+		   << lex_cast_qstring(*e->args[i]->tok) << ";";
       c_assign (t.qname(), e->args[i], "function actual argument evaluation");
     }
 
   o->newline();
   o->newline() << "if (unlikely (c->nesting+2 >= MAXNESTING)) {";
   o->newline(1) << "c->last_error = \"MAXNESTING exceeded\";";
-  o->newline(-1) << "} else {";
+  o->newline() << "c->last_stmt = " << lex_cast_qstring(*e->tok) << ";";
+  o->newline(-1) << "} else if (likely (! c->last_error)) {";
   o->indent(1);
 
   // copy in actual arguments
@@ -2263,12 +2360,12 @@ translate_pass (systemtap_session& s)
       s.op->newline() << "#if TEST_MODE";
       s.op->newline() << "#include \"runtime.h\"";
       s.op->newline() << "#else";
+      s.op->newline() << "#define STP_NUM_STRINGS 1";
       s.op->newline() << "#include \"runtime.h\"";
+      s.op->newline() << "#include \"current.c\"";
+      s.op->newline() << "#include \"stack.c\"";
       s.op->newline() << "#include <linux/string.h>";
       s.op->newline() << "#include <linux/timer.h>";
-      // XXX
-      s.op->newline() << "#define KALLSYMS_LOOKUP_NAME \"\"";
-      s.op->newline() << "#define KALLSYMS_LOOKUP 0";
       s.op->newline() << "#endif";
 
       s.up->emit_common_header ();
@@ -2326,10 +2423,7 @@ translate_pass (systemtap_session& s)
       s.op->newline(-1) << "}";
       s.op->newline();
       s.op->newline() << "void probe_exit () {";
-      // XXX: need to reference these static functions for -Werror avoidance
-      s.op->newline(1) << "if (0) next_fmt ((void *) 0, (void *) 0);";
-      s.op->newline() << "if (0) _stp_dbug(\"\", 0, \"\");";
-      s.op->newline() << "systemtap_module_exit ();";
+      s.op->newline(1) << "systemtap_module_exit ();";
       s.op->newline(-1) << "}";
 
       s.op->newline() << "MODULE_DESCRIPTION(\"systemtap probe\");";
